#ifndef INPUT_HPP
#define INPUT_HPP

#include <exception>
#include <iomanip>
#include <sstream>
#include <string>
#include <type_traits>

#include <SDL2/SDL_events.h>
#include <SDL2/SDL_keyboard.h>
#include <SDL2/SDL_render.h>

#include <basic_widgets/abstract/keyboard_handler.hpp>
#include <basic_widgets/abstract/mouse_handler.hpp>
#include <basic_widgets/caption.hpp>
#include <basic_widgets/utils/math.hpp>

namespace bwidgets::abstract
{
    template<typename T>
        concept CanToString = requires(T value) { std::to_string(value); };

    template<typename T>
        concept Printable = requires(T value) { std::declval<std::ostream&>() << value; };

    template<typename T> class Input : public Widget,
                                       public FontHandler,
                                       public KeyboardHandler,
                                       public MouseHandler
    {
        protected:
            const int       _border_width  {3};
            SDL_Rect        _input_area    {};
            widget::Caption _input_caption;
            T               _value         {};

            Input(Widget* parent=nullptr)
                : Widget(parent)
            {
                FocusHandler::_focus_area = &_input_area;
                MouseHandler::_click_area = &_input_area;
                _input_caption.text(value_to_string(_value));
            }

            virtual void _handle_focus_change()
            {
                if (FocusHandler::_has_focus)
                {
                    SDL_StartTextInput();
                }
                else
                {
                    value(value_from_string(input_text()));
                    input_text(value_to_string(_value));
                    SDL_StopTextInput();
                }
            }

            virtual void _handle_font_change()
            {
                _input_caption.font(_font);
                _update_widget_area();
            }

            virtual void _handle_key(const SDL_KeyboardEvent& key)
            {
                if (key.type == SDL_KEYDOWN)
                {
                    switch (key.keysym.sym)
                    {
                        case SDLK_BACKSPACE:
                            {
                                std::string txt = input_text();
                                if (txt.length() > 0)
                                {
                                    txt.pop_back();
                                    input_text(txt);
                                }
                                break;
                            }
                        case SDLK_RETURN:
                        case SDLK_RETURN2: // what is return2 btw?
                        case SDLK_KP_ENTER:
                            value(value_from_string(input_text()));
                            input_text(value_to_string(_value));
                            break;
                    }
                }
            }

            virtual void _handle_text_input(const SDL_TextInputEvent& input)
            {
                if (is_valid_input(input.text))
                {
                    std::string new_txt_value = input_text() + input.text;
                    input_text(new_txt_value);
                }
            }

            virtual void _update_widget_area()
            {
                _widget_area = {
                    0,
                    utils::math::center_rect(_viewport.h,
                            _input_caption.size().h + 2 * _border_width),
                    _viewport.w,
                    _input_caption.size().h + 2 * _border_width
                };
                _input_area = {
                    _widget_area.x + _border_width,
                    _widget_area.y + _border_width,
                    _widget_area.w - 2 * _border_width,
                    _widget_area.h - 2 * _border_width
                };
            }

        public:
            SDL_Color color_border     {160, 160, 160, SDL_ALPHA_OPAQUE};
            SDL_Color color_bg         {200, 200, 200, SDL_ALPHA_OPAQUE};
            int       float_precision  {2};
            SDL_Color color_bg_focused {255, 255, 255, SDL_ALPHA_OPAQUE};

            virtual void color_fg(const SDL_Color& c)
            {
                _input_caption.color_fg(c);
            }

            virtual const std::string& input_text() const
            {
                return _input_caption.text();
            }

            virtual void input_text(std::string txt)
            {
                _input_caption.text(txt);
            }

            virtual bool is_valid_input(const std::string) const
            {
                return true;
            }

            virtual T process_value(T x) const
            {
                return x;
            }

            virtual void render(SDL_Renderer* r)
            {
                Widget::render(r);

                SDL_RenderSetViewport(r, &_viewport);

                // TODO: smoothstep
                for (int i = _border_width; i > 0; i--)
                {
                    SDL_SetRenderDrawColor(
                            r,
                            color_border.r / i,
                            color_border.g / i,
                            color_border.b / i,
                            color_border.a
                            );
                    SDL_Rect border_area = {
                        _input_area.x - i,
                        _input_area.y - i,
                        _input_area.w + 2 * i,
                        _input_area.h + 2 * i
                    };
                    SDL_RenderDrawRect(r, &border_area);
                }

                if (MouseHandler::_is_hovered || FocusHandler::_has_focus)
                    _input_caption.color_bg = color_bg_focused;
                else
                    _input_caption.color_bg = color_bg;

                _input_caption.render(r);
            }

            T value() const
            {
                return _value;
            }

            void value(T new_value)
            {
                _value = new_value;
            }

            T value_from_string(std::string s)
            {
                T v;

                if constexpr(std::is_arithmetic_v<T>)
                {
                    s = s.length() > 0 ? s : "0";
                    try
                    {
                        if constexpr(std::is_floating_point_v<T>)
                            v = (T)std::stold(s);
                        else if constexpr(std::is_integral_v<T>)
                            v = (T)std::stoll(s);
                    }
                    catch(std::exception&)
                    {
                        v = value();
                    }
                }
                else if constexpr(std::is_same_v<T, std::string> || std::convertible_to<std::string, T>)
                    v = s;
                else
                    static_assert(sizeof(T) && false, "string cannot be converted to v type T.");

                return process_value(v);
            }

            std::string value_to_string(T value)
            {
                std::string s;

                if constexpr(std::is_same_v<std::string, T> || std::convertible_to<T, std::string>)
                    s = std::move(value);
                else if constexpr(CanToString<T>)
                {
                    std::stringstream ss;
                    ss << std::fixed
                        << std::setprecision(float_precision)
                        << value;
                    s = std::move(ss).str();
                }
                else if constexpr(Printable<T>)
                {
                    std::stringstream ss;
                    ss << value;
                    s = std::move(ss).str();
                }
                else
                    static_assert(sizeof(T) && false, "_value cannot be converted to string.");

                return s;
            }

            virtual void viewport(const SDL_Rect& vp)
            {
                Widget::viewport(vp);

                _input_caption.viewport(
                        utils::math::rect_offset(_input_area, _viewport)
                        );
            }
    };
}

#endif
